# Tests that read public data over the internet
import intake
import pytest
import xarray as xr


# RasterIOSource
@pytest.mark.xfail
def test_open_rasterio_http():
    prefix = 'https://landsat-pds.s3.us-west-2.amazonaws.com/L8/139/045'
    image = 'LC81390452014295LGN00/LC81390452014295LGN00_B1.TIF'
    url = f'{prefix}/{image}'
    source = intake.open_rasterio(url,
                                  chunks=dict(band=1))
    ds = source.to_dask()
    assert isinstance(ds, xr.core.dataarray.DataArray)


def test_open_rasterio_s3():
    bucket = 's3://landsat-pds'
    key = 'L8/139/045/LC81390452014295LGN00/LC81390452014295LGN00_B1.TIF'
    url = f'{bucket}/{key}'
    source = intake.open_rasterio(url,
                                  chunks=dict(band=1),
                                  storage_options = dict(anon=True))
    ds = source.to_dask()
    assert isinstance(ds, xr.core.dataarray.DataArray)


# NETCDFSource
def test_open_netcdf_gs():
    bucket = 'gs://ldeo-glaciology'
    key = 'bedmachine/BedMachineAntarctica_2019-11-05_v01.nc'
    url = f'{bucket}/{key}'
    source = intake.open_netcdf(url,
                                chunks=3000,
                                xarray_kwargs=dict(engine='h5netcdf'),
                                )
    ds = source.to_dask()
    assert isinstance(ds, xr.core.dataarray.Dataset)


@pytest.mark.xfail
def test_open_netcdf_s3():
    bucket = 's3://its-live-data.jpl.nasa.gov'
    key = 'icesat2/alt06/rel003/ATL06_20181230162257_00340206_003_01.h5'
    url = f'{bucket}/{key}'
    source = intake.open_netcdf(url,
                                xarray_kwargs=dict(group='gt1l/land_ice_segments', engine='h5netcdf'),
                                storage_options=dict(anon=True),
                                )
    ds = source.to_dask()
    assert isinstance(ds._file_obj, xr.backends.h5netcdf_.H5NetCDFStore)
    assert isinstance(ds, xr.core.dataarray.Dataset)


@pytest.mark.xfail
def test_open_netcdf_s3_simplecache():
    bucket = 's3://its-live-data.jpl.nasa.gov'
    key = 'icesat2/alt06/rel003/ATL06_20181230162257_00340206_003_01.h5'
    url = f'simplecache::{bucket}/{key}'
    source = intake.open_netcdf(url,
                                xarray_kwargs=dict(group='gt1l/land_ice_segments', engine='h5netcdf'),
                                storage_options=dict(s3={'anon': True}),
                                )
    ds = source.to_dask()
    assert isinstance(ds._file_obj, xr.backends.h5netcdf_.H5NetCDFStore)
    assert isinstance(ds, xr.core.dataarray.Dataset)


def test_open_opendap():
    url = 'https://www.ncei.noaa.gov/thredds/dodsC/model-cbofs-files/2021/12/nos.cbofs.fields.n001.20211231.t18z.nc'

    source = intake.open_opendap(url, engine='netcdf4', chunks={'time': 1})
    ds = source.to_dask()
    assert isinstance(ds, xr.core.dataarray.Dataset)


def test_open_list_opendap():
    url1 = 'https://www.ncei.noaa.gov/thredds/dodsC/model-cbofs-files/2021/12/nos.cbofs.fields.n001.20211231.t18z.nc'
    url2 = 'https://www.ncei.noaa.gov/thredds/dodsC/model-cbofs-files/2021/12/nos.cbofs.fields.n002.20211231.t18z.nc'

    source = intake.open_opendap([url1, url2], engine='netcdf4', chunks={'time': 1})
    ds = source.to_dask()
    assert isinstance(ds, xr.core.dataarray.Dataset)



# -*- coding: utf-8 -*-
import numpy as np
import os
import pytest

from intake import open_catalog


@pytest.fixture
def catalog1():
    path = os.path.dirname(__file__)
    return open_catalog(os.path.join(path, 'data', 'catalog.yaml'))


def test_catalog(catalog1, dataset):
    source = catalog1['xarray_source'].get()
    ds = source.read()

    assert ds.dims == dataset.dims
    assert np.all(ds.temp == dataset.temp)
    assert np.all(ds.rh == dataset.rh)


def test_persist(catalog1):
    from intake_xarray import ZarrSource
    source = catalog1['blank']
    s2 = source.persist()
    assert source.has_been_persisted
    assert isinstance(s2, ZarrSource)
    assert s2.is_persisted
    assert (source.read() == s2.read()).all()


def test_import_error(mock_import_xarray, catalog1):
    s = catalog1['xarray_source']()  # this is OK
    with pytest.raises(ImportError):
        s.discover()



from . import __version__
from intake.source.base import DataSource, Schema


class DataSourceMixin(DataSource):
    """Common behaviours for plugins in this repo"""
    version = __version__
    container = 'xarray'
    partition_access = True

    def _get_schema(self):
        """Make schema object, which embeds xarray object and some details"""
        from .xarray_container import serialize_zarr_ds

        self.urlpath = self._get_cache(self.urlpath)[0]

        if self._ds is None:
            self._open_dataset()

            metadata = {
                'dims': dict(self._ds.dims),
                'data_vars': {k: list(self._ds[k].coords)
                              for k in self._ds.data_vars.keys()},
                'coords': tuple(self._ds.coords.keys()),
            }
            if getattr(self, 'on_server', False):
                metadata['internal'] = serialize_zarr_ds(self._ds)
            metadata.update(self._ds.attrs)
            self._schema = Schema(
                datashape=None,
                dtype=None,
                shape=None,
                npartitions=None,
                extra_metadata=metadata)
        return self._schema

    def read(self):
        """Return a version of the xarray with all the data in memory"""
        self._load_metadata()
        return self._ds.load()

    def read_chunked(self):
        """Return xarray object (which will have chunks)"""
        self._load_metadata()
        return self._ds

    def read_partition(self, i):
        """Fetch one chunk of data at tuple index i
        """
        import numpy as np
        self._load_metadata()
        if not isinstance(i, (tuple, list)):
            raise TypeError('For Xarray sources, must specify partition as '
                            'tuple')
        if isinstance(i, list):
            i = tuple(i)
        if hasattr(self._ds, 'variables') or i[0] in self._ds.coords:
            arr = self._ds[i[0]].data
            i = i[1:]
        else:
            arr = self._ds.data
        if isinstance(arr, np.ndarray):
            return arr
        # dask array
        return arr.blocks[i].compute()

    def to_dask(self):
        """Return xarray object where variables are dask arrays"""
        return self.read_chunked()

    def close(self):
        """Delete open file from memory"""
        self._ds = None
        self._schema = None



from intake import Schema
from intake.source.derived import GenericTransform


class XArrayTransform(GenericTransform):
    """Transform where the input and output are both xarray objects.
    You must supply ``transform`` and any ``transform_kwargs``.
    """

    input_container = "xarray"
    container = "xarray"
    optional_params = {}
    _ds = None

    def to_dask(self):
        if self._ds is None:
            self._pick()
            self._ds = self._transform(
                self._source.to_dask(), **self._params["transform_kwargs"]
            )
        return self._ds

    def _get_schema(self):
        """load metadata only if needed"""
        self.to_dask()
        return Schema(
            datashape=None,
            dtype=None,
            shape=None,
            npartitions=None,
            extra_metadata=self._ds.extra_metadata,
        )

    def read(self):
        return self.to_dask().compute()


class Sel(XArrayTransform):
    """Simple array transform to subsample an xarray object using
    the sel method.
    Note that you could use XArrayTransform directly, by writing a
    function to choose the subsample instead of a method as here.
    """

    input_container = "xarray"
    container = "xarray"
    required_params = ["indexers"]

    def __init__(self, indexers, **kwargs):
        """
        indexers: dict (stord as str) which is passed to xarray.Dataset.sel
        """
        # this class wants required "indexers", but XArrayTransform
        # uses "transform_kwargs", which we don't need since we use a method for the
        # transform
        kwargs.update(
            transform=self.sel,
            indexers=indexers,
            transform_kwargs={},
        )
        super().__init__(**kwargs)

    def sel(self, ds):
        return ds.sel(eval(self._params["indexers"]))



# -*- coding: utf-8 -*-
import os
import numpy as np
import pytest

from intake_xarray.image import _coerce_shape, ImageSource

here = os.path.dirname(__file__)


@pytest.mark.parametrize('im', [
    [[1, 2],
     [3, 4]],
    [[1, 2, 7],
     [3, 4, 6]],
    [[1, 2, 7],
     [3, 4, 6],
     [5, 6, 8]],
    [[1, 2],
     [3, 4],
     [5, 6],
     [7, 8]],
])
def test_coerce_shape_2d_trim_only(im):
    shape = (2, 2)
    array = np.array(im)
    expected = np.array([[1, 2],
                         [3, 4]])
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()


def test_coerce_shape_2d_pad_only():
    shape = (3, 4)
    array = np.array([[1, 2],
                      [3, 4]])
    expected = np.array([[1, 2, 0, 0],
                         [3, 4, 0, 0],
                         [0, 0, 0, 0]])
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()


def test_coerce_shape_2d_pad_nrows_and_trim_ncols():
    shape = (4, 2)
    array = np.array([[1, 2, 7],
                      [3, 4, 6]])
    expected = np.array([[1, 2],
                         [3, 4],
                         [0, 0],
                         [0, 0]])
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()


def test_coerce_shape_2d_pad_ncols_and_trim_nrows():
    shape = (2, 4)
    array = np.array([[1, 2],
                      [3, 4],
                      [5, 6],
                      [7, 8]])
    expected = np.array([[1, 2, 0, 0],
                         [3, 4, 0, 0]])
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()


def test_coerce_shape_3d_no_change():
    shape = (3, 3)
    array = np.arange(3**3).reshape(3, 3, 3)
    actual = _coerce_shape(array, shape)
    assert (array == actual).all()


def test_coerce_shape_3d_pad_nrows_and_trim_ncols():
    shape = (5, 2)
    array = np.arange(2*4*3).reshape(2, 4, 3)
    expected = np.array([[[0,  1,  2],
                          [3,  4,  5]],

                         [[12, 13, 14],
                          [15, 16, 17]],

                         [[0,  0,  0],
                          [0,  0,  0]],

                         [[0,  0,  0],
                          [0,  0,  0]],

                         [[0,  0,  0],
                          [0,  0,  0]]])
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()


def test_coerce_shape_3d_pad_ncols_and_trim_nrows():
    shape = (2, 5)
    array = np.arange(3*2*4).reshape(3, 2, 4)
    expected = np.array([[[0,  1,  2,  3],
                          [4,  5,  6,  7],
                          [0,  0,  0,  0],
                          [0,  0,  0,  0],
                          [0,  0,  0,  0]],

                         [[8,  9, 10, 11],
                          [12, 13, 14, 15],
                          [0,  0,  0,  0],
                          [0,  0,  0,  0],
                          [0,  0,  0,  0]]])
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()


def test_coerce_shape_raises_error_if_shape_not_len_2():
    shape = (2, 3, 3)
    array = np.arange(3**3).reshape(3, 3, 3)
    with pytest.raises(ValueError,
                       match='coerce_shape must be an iterable of len 2'):
        _coerce_shape(array, shape)


def test_coerce_shape_array_non_int():
    shape = (2, 3)
    array = np.random.random((3, 2))
    expected = np.append(array[:2, :], [[0], [0]], axis=1)
    actual = _coerce_shape(array, shape)
    assert (expected == actual).all()
    assert expected.dtype == "float"


def test_read_image():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', 'beach57.tif')
    source = ImageSource(urlpath=urlpath)
    array = source.read()
    assert array.shape == (256, 252, 3)
    assert array.dtype == np.uint8


def test_read_image_and_exif():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', 'beach57.tif')
    source = ImageSource(urlpath=urlpath, exif_tags=True)
    ds = source.read()
    assert ds['raster'].shape == (256, 252, 3)
    assert ds['raster'].dtype == np.uint8
    assert ds['EXIF Image ImageWidth'].item().values == [252]
    assert ds['EXIF Image ImageLength'].item().values == [256]


def test_read_image_and_given_exif_tag():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', 'beach57.tif')
    source = ImageSource(urlpath=urlpath, exif_tags=['Image ImageWidth'])
    ds = source.read()
    assert ds['raster'].shape == (256, 252, 3)
    assert ds['raster'].dtype == np.uint8
    assert ds['EXIF Image ImageWidth'].item().values == [252]
    with pytest.raises(KeyError):
        ds['EXIF Image ImageLength']


def test_read_images_as_glob_without_coerce_raises_error():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', '*')
    source = ImageSource(urlpath=urlpath)
    with pytest.raises(ValueError,
                       match='could not broadcast input array'):
        source.read()


def test_read_images_as_glob_with_coerce():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', '*')
    source = ImageSource(urlpath=urlpath, coerce_shape=(256, 256))
    array = source.read()
    assert array.shape == (3, 256, 256, 3)


def test_read_images_and_exif_as_glob_with_coerce():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', '*')
    source = ImageSource(urlpath=urlpath, coerce_shape=(256, 256), exif_tags=True)
    ds = source.read()
    assert ds['raster'].shape == (3, 256, 256, 3)
    assert ds['EXIF Image ImageWidth'].shape == (3,)


def test_read_images_and_persist():
    pytest.importorskip('skimage')
    urlpath = os.path.join(here, 'data', 'images', '*')
    source = ImageSource(urlpath=urlpath, coerce_shape=(256, 256))
    import tempfile
    exported = tempfile.mkdtemp()
    source.export(exported)
    import xarray as xr
    assert xr.open_dataset(exported, engine="zarr")
